namespace Gate {

  public enum StateCode { Close, Open, Unknown }
  public enum EventCode { Coin, Pass, Unknown }

  public interface Context {
    void Error(StateCode s, EventCode e);
    bool Confirm();
    void Lock();
    void Unlock();
    void Alarm();
    void ThankYou();
  }

  public abstract class State {
    public abstract string Name { get; }
    public abstract StateCode Code { get; }
    internal virtual void Enter_(Context ctx) {}
    internal virtual void Exit_(Context ctx) {}
    internal virtual State Coin(Context ctx) 
      { ctx.Error(Code,EventCode.Coin); return this; }
    internal virtual State Pass(Context ctx) 
      { ctx.Error(Code,EventCode.Pass); return this; }
  }

  internal class CloseState : State {
    public override string Name { get { return "Close"; } }
    public override StateCode Code { get { return StateCode.Close; } }
    internal override void Enter_(Context ctx) { ctx.Lock(); }
    internal override State Coin(Context ctx) {
      if ( ctx.Confirm() ) {
        Exit_(ctx);
        StateMachine.Open.Enter_(ctx);
        return StateMachine.Open;
      }
      return this;
    }
    internal override State Pass(Context ctx) {
      ctx.Alarm();
      return this;
    }
  }

  internal class OpenState : State {
    public override string Name { get { return "Open"; } }
    public override StateCode Code { get { return StateCode.Open; } }
    internal override void Enter_(Context ctx) { ctx.Unlock(); }
    internal override State Coin(Context ctx) {
      if ( !ctx.Confirm() ) return this;
      ctx.ThankYou();
      return this;
    }
    internal override State Pass(Context ctx) {
      Exit_(ctx);
      StateMachine.Close.Enter_(ctx);
      return StateMachine.Close;
    }
  }

  public class StateMachine {
    internal static State Close = new CloseState();
    internal static State Open = new OpenState();
    private State initial_;
    private State state_;
    public StateMachine() { initial_ = Close; }
    public State CurrentState { get { return state_; } }
    public void Start(Context ctx) { state_ = initial_; state_.Enter_(ctx); }
    public void Coin(Context ctx) { state_ = state_.Coin(ctx); }
    public void Pass(Context ctx) { state_ = state_.Pass(ctx); }
  }
}
