using System;
using System.IO;
using System.Collections.Generic;
using System.Xml;
using System.Diagnostics;

namespace smc {

  class StateMapCompiler {

    // listvaluedȂǉB
    static bool addUnique(IList<string> list, string value) {
      if ( !list.Contains(value) ) {
        list.Add(value);
        return true;
      }
      return false;
    }

    static void Main(string[] args) {

      XmlDocument document = new XmlDocument();
      document.PreserveWhitespace = false;
      document.Load(args[0]);
      XmlElement root = document.DocumentElement;
      XmlAttribute attribute;

      // state
      Debug.Write("state : ");
      List<string> state_names = new List<string>();
      foreach ( XmlNode node in root.GetElementsByTagName("state") ) {
        attribute = node.Attributes["name"];
        if ( addUnique(state_names, attribute.Value) ) {
          Debug.Write(string.Format("{0} ", attribute.Value));
        }
      }
      Debug.WriteLine("");

      // event
      Debug.Write("event : ");
      List<string> event_names = new List<string>();
      foreach ( XmlNode node in root.GetElementsByTagName("event") ) {
        attribute = node.Attributes["name"];
        if ( addUnique(event_names, attribute.Value) ) {
          Debug.Write(string.Format("{0} ", node.Attributes["name"].Value));
        }
      }
      Debug.WriteLine("");

      // predicate
      Debug.Write("predicate : ");
      List<string> predicate_names = new List<string>();
      foreach ( XmlNode node in root.GetElementsByTagName("transit") ) {
        attribute = node.Attributes["if"];
        if ( attribute != null ) {
          if ( addUnique( predicate_names, attribute.Value ) ) {
            Debug.Write(string.Format("{0} ", attribute.Value));
          }
        }
      }
      foreach ( XmlNode node in root.GetElementsByTagName("event") ) {
        attribute = node.Attributes["guard"];
        if ( attribute != null ) {
          if ( addUnique( predicate_names, attribute.Value ) ) {
            Debug.Write(string.Format("{0} ", attribute.Value));
          }
        }
      }
      Debug.WriteLine("");

      // action
      Debug.Write("action : ");
      List<string> action_names = new List<string>();
      foreach ( XmlNode node in root.GetElementsByTagName("state") ) {
        attribute = node.Attributes["entry"];
        if ( attribute != null ) {
          if ( addUnique( action_names, attribute.Value) ) {
            Debug.Write(string.Format("{0} ", attribute.Value));
          }
        }
        attribute = node.Attributes["exit"];
        if ( attribute != null ) {
          if ( addUnique( action_names, attribute.Value) ) {
            Debug.Write(string.Format("{0} ", attribute.Value));
          }
        }
      }
      foreach ( XmlNode node in root.GetElementsByTagName("execute") ) {
        attribute = node.Attributes["action"];
        if ( addUnique( action_names, attribute.Value) ) {
          Debug.Write(string.Format("{0} ", attribute.Value));
        }
      }
      Debug.WriteLine("");

      /*
       * code generation starts here!
       */

      String path = args[0];
      int pos = path.LastIndexOf('.');
      if ( pos >= 0 ) path = path.Substring(0,pos);
      path += ".cs";
      StreamWriter writer = new StreamWriter(path);

      // ----- namespace
      writer.Write(          "namespace {0} {{\n\n", root.Attributes["name"].Value);

      // ----- enum StateCode
      writer.Write(          "  public enum StateCode { ");
      foreach ( string value in state_names ) {
        writer.Write("{0}, ", value);
      }
      writer.Write("Unknown }\n");

      // ----- enum EventCode
      writer.Write(          "  public enum EventCode { ");
      foreach ( string value in event_names ) {
        writer.Write("{0}, ", value);
      }
      writer.WriteLine("Unknown }\n");

      // ----- interface Context
      writer.Write(          "  public interface Context {\n" + 
                             "    void Error(StateCode s, EventCode e);\n");
      foreach ( string value in predicate_names ) {
        writer.Write(        "    bool {0}();\n", value);
      }
      foreach ( string value in action_names ) {
        writer.Write(        "    void {0}();\n", value);
      }
      writer.Write(          "  }\n\n");

      // ----- abstract class State
      writer.Write(          "  public abstract class State {\n" +
                             "    public abstract string Name { get; }\n" +
                             "    public abstract StateCode Code { get; }\n" +
                             "    internal virtual void Enter_(Context ctx) {}\n" +
                             "    internal virtual void Exit_(Context ctx) {}\n");
      foreach ( string value in event_names ) {
        writer.Write(        "    internal virtual State {0}(Context ctx) \n" +
                             "      {{ ctx.Error(Code,EventCode.{0}); return this; }}\n", 
                             value);
      }
      writer.Write(          "  }\n\n");

      // ----- class xxxState : State
      foreach ( XmlNode node in root.GetElementsByTagName("state") ) {
        attribute = node.Attributes["name"];
        writer.Write(        "  internal class {0}State : State {{\n" +
                             "    public override string Name {{ get {{ return \"{0}\"; }} }}\n" +
                             "    public override StateCode Code {{ get {{ return StateCode.{0}; }} }}\n",
                             attribute.Value);
        attribute = node.Attributes["entry"];
        if ( attribute != null ) {
          writer.Write(      "    internal override void Enter_(Context ctx) {{ ctx.{0}(); }}\n", 
                             attribute.Value);
        }
        attribute = node.Attributes["exit"];
        if ( attribute != null ) {
          writer.Write(      "    internal override void Exit_(Context ctx) {{ ctx.{0}(); }}\n", 
                             attribute.Value);
        }
        foreach ( XmlNode event_node in node.ChildNodes ) {
          bool default_return = true;
          writer.Write(      "    internal override State {0}(Context ctx) {{\n", 
                             event_node.Attributes["name"].Value);
          if ( event_node.Attributes["guard"] != null ) {
            writer.Write(    "      if ( !ctx.{0}() ) return this;\n", 
                             event_node.Attributes["guard"].Value);
          }
          foreach ( XmlNode action_node in event_node ) {
            if ( action_node.Name == "execute" ) {
              writer.Write(  "      ctx.{0}();\n", action_node.Attributes["action"].Value);
            }
            if ( action_node.Name == "transit" ) {
              attribute = action_node.Attributes["if"];
              if ( attribute != null ) {
                writer.Write("      if ( ctx.{0}() ) {{\n" +
                             "        Exit_(ctx);\n" +
                             "        StateMachine.{1}.Enter_(ctx);\n" +
                             "        return StateMachine.{1};\n" +
                             "      }}\n", 
                             attribute.Value, action_node.Attributes["state"].Value);
              } else {
                writer.Write("      Exit_(ctx);\n" +
                             "      StateMachine.{0}.Enter_(ctx);\n" +
                             "      return StateMachine.{0};\n", 
                             action_node.Attributes["state"].Value);
                default_return = false;
                break;
              }
            }
          }
          if ( default_return ) {
            writer.Write(    "      return this;\n");
          }
          writer.Write(      "    }\n");
        }
        writer.Write(        "  }\n\n");
      }

      // ----- class StateMachine
      writer.Write(          "  public class StateMachine {\n");
      foreach ( string value in  state_names ) {
        writer.Write(        "    internal static State {0} = new {0}State();\n", value);
      }
      writer.Write(          "    private State initial_;\n" +
                             "    private State state_;\n" +
                             "    public StateMachine() {{ initial_ = {0}; }}\n" +
                             "    public State CurrentState {{ get {{ return state_; }} }}\n" +
                             "    public void Start(Context ctx) {{ state_ = initial_; state_.Enter_(ctx); }}\n",
                             root.Attributes["initial"].Value);
      foreach ( string value in event_names ) {
        writer.Write(        "    public void {0}(Context ctx) {{ state_ = state_.{0}(ctx); }}\n", value);
      }
      writer.Write(          "  }\n" +
                             "}\n");
      writer.Close();

    }

  }
}
